@@ -22,6 +22,7 @@
 #include "KCupsConnection.h"
 
 #include "KCupsPasswordDialog.h"
+#include "KIppRequest.h"
 
 #include <config.h>
 
@@ -68,20 +69,6 @@ static int password_retries = 0;
 static int internalErrorCount = 0;
 const char * password_cb(const char *prompt, http_t *http, const char *method, const char *resource, void *user_data);
 
-static const char **qStringListToCharPtrPtr(const QStringList &list, QList<QByteArray> *qbaList)
-{
-    const char **ptr = new const char *[list.size() + 1];
-    qbaList->reserve(qbaList->size() + list.size());
-    QByteArray qba;
-    for (int i = 0; i < list.size(); ++i) {
-        qba = list.at(i).toUtf8();
-        qbaList->append(qba);
-        ptr[i] = qba.constData();
-    }
-    ptr[list.size()] = 0;
-    return ptr;
-}
-
 KCupsConnection* KCupsConnection::global()
 {
     if (!m_instance) {
@@ -101,6 +88,7 @@ KCupsConnection::KCupsConnection(const KUrl &server, QObject *parent) :
     QThread(parent),
     m_serverUrl(server)
 {
+    qRegisterMetaType<KIppRequest>("KIppRequest");
     init();
 }
 
@@ -118,6 +106,11 @@ KCupsConnection::~KCupsConnection()
     delete m_subscriptionTimer;
 }
 
+void KCupsConnection::setPasswordMainWindow(WId mainwindow)
+{
+    m_passwordDialog->setMainWindow(mainwindow);
+}
+
 void KCupsConnection::init()
 {
     // Creating the dialog before start() will make it run on the gui thread
@@ -325,69 +318,20 @@ bool KCupsConnection::readyToStart()
     return false;
 }
 
-ReturnArguments KCupsConnection::request(ipp_op_e operation,
-                                         const char *resource,
-                                         const QVariantHash &reqValues,
-                                         bool needResponse)
+
+ReturnArguments KCupsConnection::request(const KIppRequest &request, ipp_tag_t groupTag) const
 {
     ReturnArguments ret;
-
-    if (!readyToStart()) {
-        return ret; // This is not intended to be used in the gui thread
-    }
-
     ipp_t *response = NULL;
-    bool needDestName = false;
-    int group_tag = IPP_TAG_PRINTER;
     do {
-        ipp_t *request;
-        bool isClass = false;
-        QString filename;
-        QVariantHash values = reqValues;
-
         ippDelete(response);
+        response = NULL;
 
-        if (values.contains(QLatin1String("printer-is-class"))) {
-            isClass = values.take(QLatin1String("printer-is-class")).toBool();
-        }
-        if (values.contains(QLatin1String("need-dest-name"))) {
-            needDestName = values.take(QLatin1String("need-dest-name")).toBool();
-        }
-        if (values.contains(QLatin1String("group-tag-qt"))) {
-            group_tag = values.take(QLatin1String("group-tag-qt")).toInt();
-        }
+        response = request.sendIppRequest();
+    } while (retry(request.resource().toUtf8(), request.operation()));
 
-        if (values.contains(QLatin1String("filename"))) {
-            filename = values.take(QLatin1String("filename")).toString();
-        }
-
-        // Lets create the request
-        if (values.contains(QLatin1String(KCUPS_PRINTER_NAME))) {
-            request = ippNewDefaultRequest(values.take(QLatin1String(KCUPS_PRINTER_NAME)).toString(),
-                                           isClass,
-                                           operation);
-        } else {
-            request = ippNewRequest(operation);
-        }
-
-        // send our user name on the request too
-        ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_NAME,
-                     "requesting-user-name", NULL, cupsUser());
-
-        // Add the requested values to the request
-        requestAddValues(request, values);
-
-        // Do the request
-        // do the request deleting the response
-        if (filename.isEmpty()) {
-            response = cupsDoRequest(CUPS_HTTP_DEFAULT, request, resource);
-        } else {
-            response = cupsDoFileRequest(CUPS_HTTP_DEFAULT, request, resource, filename.toUtf8());
-        }
-    } while (retry(resource));
-
-    if (response != NULL && needResponse) {
-        ret = parseIPPVars(response, group_tag, needDestName);
+    if (response && groupTag != IPP_TAG_ZERO) {
+        ret = parseIPPVars(response, groupTag, true);
     }
     ippDelete(response);
 
@@ -398,53 +342,40 @@ int KCupsConnection::renewDBusSubscription(int subscriptionId, int leaseDuration
 {
     int ret = -1;
 
-    if (!readyToStart()) {
-        kWarning() << "Tryied to run on the wrong thread";
-        return subscriptionId; // This is not intended to be used in the gui thread
+    ipp_op_t operation;
+
+    // check if we have a valid subscription ID
+    if (subscriptionId >= 0) {
+        // Add the "notify-events" values to the request
+        operation = IPP_RENEW_SUBSCRIPTION;
+    } else {
+        operation = IPP_CREATE_PRINTER_SUBSCRIPTION;
+    }
+
+    KIppRequest request(operation, "/");
+    request.addString(IPP_TAG_OPERATION, IPP_TAG_URI,
+                       KCUPS_PRINTER_URI, QLatin1String("/"));
+    request.addInteger(IPP_TAG_SUBSCRIPTION, IPP_TAG_INTEGER,
+                        KCUPS_NOTIFY_LEASE_DURATION, leaseDuration);
+
+    if (operation == IPP_CREATE_PRINTER_SUBSCRIPTION) {
+        // Add the "notify-events" values to the request
+        request.addStringList(IPP_TAG_SUBSCRIPTION, IPP_TAG_KEYWORD,
+                              KCUPS_NOTIFY_EVENTS, events);
+        request.addString(IPP_TAG_SUBSCRIPTION, IPP_TAG_KEYWORD,
+                          KCUPS_NOTIFY_PULL_METHOD, "ippget");
+        request.addString(IPP_TAG_SUBSCRIPTION, IPP_TAG_URI,
+                          KCUPS_NOTIFY_RECIPIENT_URI, "dbus://");
+    } else {
+        request.addInteger(IPP_TAG_OPERATION, IPP_TAG_INTEGER,
+                           KCUPS_NOTIFY_SUBSCRIPTION_ID, subscriptionId);
     }
 
     ipp_t *response = NULL;
     do {
-        ipp_t *request;
-        ipp_op_e operation;
-
-        // check if we have a valid subscription ID
-        if (subscriptionId >= 0) {
-            // Add the "notify-events" values to the request
-            operation = IPP_RENEW_SUBSCRIPTION;
-        } else {
-            operation = IPP_CREATE_PRINTER_SUBSCRIPTION;
-        }
-
-        // Lets create the request
-        request = ippNewRequest(operation);
-        ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI,
-                     KCUPS_PRINTER_URI, NULL, "/");
-        ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_NAME,
-                     "requesting-user-name", NULL, cupsUser());
-
-        if (operation == IPP_CREATE_PRINTER_SUBSCRIPTION) {
-            // Add the "notify-events" values to the request
-            QVariantHash values;
-            values["notify-events"] = events;
-            requestAddValues(request, values);
-
-            ippAddString(request, IPP_TAG_SUBSCRIPTION, IPP_TAG_KEYWORD,
-                         "notify-pull-method", NULL, "ippget");
-            ippAddString(request, IPP_TAG_SUBSCRIPTION, IPP_TAG_URI,
-                         "notify-recipient-uri", NULL, "dbus://");
-            ippAddInteger(request, IPP_TAG_SUBSCRIPTION, IPP_TAG_INTEGER,
-                          "notify-lease-duration", leaseDuration);
-        } else {
-            ippAddInteger(request, IPP_TAG_OPERATION, IPP_TAG_INTEGER,
-                          "notify-subscription-id", subscriptionId);
-            ippAddInteger(request, IPP_TAG_SUBSCRIPTION, IPP_TAG_INTEGER,
-                          "notify-lease-duration", leaseDuration);
-        }
-
         // Do the request
-        response = cupsDoRequest(CUPS_HTTP_DEFAULT, request, "/");
-    } while (retry("/"));
+        response = request.sendIppRequest();
+    } while (retry("/", operation));
 
 #if CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR >= 6
     if (response && ippGetStatusCode(response) == IPP_OK) {
@@ -631,125 +562,22 @@ void KCupsConnection::renewDBusSubscription()
 
 void KCupsConnection::cancelDBusSubscription()
 {
+    KIppRequest request(IPP_CANCEL_SUBSCRIPTION, "/");
+    request.addString(IPP_TAG_OPERATION, IPP_TAG_URI,
+                      KCUPS_PRINTER_URI, "/");
+    request.addInteger(IPP_TAG_OPERATION, IPP_TAG_INTEGER,
+                       KCUPS_NOTIFY_SUBSCRIPTION_ID, m_subscriptionId);
+
     do {
-        ipp_t *request;
-
-        // Lets create the request
-        request = ippNewRequest(IPP_CANCEL_SUBSCRIPTION);
-        ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI,
-                     KCUPS_PRINTER_URI, NULL, "/");
-        ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_NAME,
-                     "requesting-user-name", NULL, cupsUser());
-        ippAddInteger(request, IPP_TAG_OPERATION, IPP_TAG_INTEGER,
-                      "notify-subscription-id", m_subscriptionId);
-
         // Do the request
-        ippDelete(cupsDoRequest(CUPS_HTTP_DEFAULT, request, "/"));
-    } while (retry("/"));
+        ippDelete(request.sendIppRequest());
+    } while (retry(request.resource().toUtf8(), request.operation()));
 
     // Reset the subscription id
     m_subscriptionId = -1;
 }
 
-void KCupsConnection::requestAddValues(ipp_t *request, const QVariantHash &values)
-{
-    QVariantHash::const_iterator i = values.constBegin();
-    while (i != values.constEnd()) {
-        QString key = i.key();
-        QVariant value = i.value();
-        switch (value.type()) {
-        case QVariant::Bool:
-            if (key == QLatin1String("printer-is-accepting-jobs")) {
-                ippAddBoolean(request, IPP_TAG_PRINTER,
-                              "printer-is-accepting-jobs", value.toBool());
-            } else {
-                ippAddBoolean(request, IPP_TAG_OPERATION,
-                              key.toUtf8(), value.toBool());
-            }
-            break;
-        case QVariant::Int:
-            if (key == QLatin1String(KCUPS_JOB_ID)) {
-                ippAddInteger(request, IPP_TAG_OPERATION, IPP_TAG_INTEGER,
-                              KCUPS_JOB_ID, value.toInt());
-            } else if (key == QLatin1String(KCUPS_PRINTER_STATE)) {
-                ippAddInteger(request, IPP_TAG_PRINTER, IPP_TAG_ENUM,
-                              KCUPS_PRINTER_STATE, IPP_PRINTER_IDLE);
-            } else {
-                ippAddInteger(request, IPP_TAG_OPERATION, IPP_TAG_ENUM,
-                              key.toUtf8(), value.toInt());
-            }
-            break;
-        case QVariant::String:
-            if (key == QLatin1String(KCUPS_DEVICE_URI)) {
-                // device uri has a different TAG
-                ippAddString(request, IPP_TAG_PRINTER, IPP_TAG_URI,
-                             KCUPS_DEVICE_URI, "utf-8", value.toString().toUtf8());
-            } else if (key == QLatin1String("job-printer-uri")) {
-                // TODO this seems broken
-                const char* dest_name = value.toString().toUtf8();
-                char  destUri[HTTP_MAX_URI];
-                httpAssembleURIf(HTTP_URI_CODING_ALL, destUri, sizeof(destUri),
-                                 "ipp", "utf-8", "localhost", ippPort(),
-                                 "/printers/%s", dest_name);
-                ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI,
-                             KCUPS_JOB_PRINTER_URI, "utf-8", value.toString().toUtf8());
-            } else if (key == QLatin1String(KCUPS_PRINTER_OP_POLICY) ||
-                       key == QLatin1String(KCUPS_PRINTER_ERROR_POLICY) ||
-                       key == QLatin1String("ppd-name")) {
-                // printer-op-policy has a different TAG
-                ippAddString(request, IPP_TAG_PRINTER, IPP_TAG_NAME,
-                            key.toUtf8(), "utf-8", value.toString().toUtf8());
-            } else if (key == QLatin1String(KCUPS_JOB_NAME)) {
-                ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_NAME,
-                             KCUPS_JOB_NAME, "utf-8", value.toString().toUtf8());
-            } else if (key == QLatin1String("which-jobs")) {
-                ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_KEYWORD,
-                             "which-jobs", "utf-8", value.toString().toUtf8());
-            } else {
-                ippAddString(request, IPP_TAG_PRINTER, IPP_TAG_TEXT,
-                             key.toUtf8(), "utf-8", value.toString().toUtf8());
-            }
-            break;
-        case QVariant::StringList:
-            {
-                QStringList list = value.value<QStringList>();
-
-                QList<QByteArray> valuesQByteArrayList;
-                const char **values = qStringListToCharPtrPtr(list, &valuesQByteArrayList);
-
-                if (key == QLatin1String(KCUPS_MEMBER_URIS)) {
-                    ippAddStrings(request, IPP_TAG_PRINTER, IPP_TAG_URI,
-                                  KCUPS_MEMBER_URIS, list.size(), "utf-8", values);
-                } else if (key == QLatin1String("requested-attributes")) {
-                    ippAddStrings(request, IPP_TAG_OPERATION, IPP_TAG_KEYWORD,
-                                  "requested-attributes", list.size(), "utf-8", values);
-                } else if (key == QLatin1String("notify-events")) {
-                    // Used for DBus notification, the values contains
-                    // what we want to watch
-                    ippAddStrings(request, IPP_TAG_SUBSCRIPTION, IPP_TAG_KEYWORD,
-                                  "notify-events", list.size(), NULL, values);
-                } else {
-                    ippAddStrings(request, IPP_TAG_PRINTER, IPP_TAG_NAME,
-                                  i.key().toUtf8(), list.size(), "utf-8", values);
-                }
-
-                // ippAddStrings deep copies everything so we can throw away the values.
-                // the QBAList and content is auto discarded when going out of scope.
-                delete [] values;
-            }
-            break;
-        case QVariant::UInt:
-            ippAddInteger(request, IPP_TAG_OPERATION, IPP_TAG_ENUM,
-                          key.toUtf8(), value.toInt());
-            break;
-        default:
-            kWarning() << "type NOT recognized! This will be ignored:" << key << "values" << value;
-        }
-        ++i;
-    }
-}
-
-ReturnArguments KCupsConnection::parseIPPVars(ipp_t *response, int group_tag, bool needDestName)
+ReturnArguments KCupsConnection::parseIPPVars(ipp_t *response, ipp_tag_t group_tag, bool needDestName)
 {
     ipp_attribute_t *attr;
     ReturnArguments ret;
@@ -849,32 +677,6 @@ ReturnArguments KCupsConnection::parseIPPVars(ipp_t *response, int group_tag, bo
     return ret;
 }
 
-// Don't forget to delete the request
-ipp_t* KCupsConnection::ippNewDefaultRequest(const QString &name, bool isClass, ipp_op_t operation)
-{
-    char  uri[HTTP_MAX_URI]; // printer URI
-    ipp_t *request;
-
-    QString destination;
-    if (isClass) {
-        destination = QLatin1String("/classes/") % name;
-    } else {
-        destination = QLatin1String("/printers/") % name;
-    }
-
-    // Create a new request
-    // where we need:
-    // * printer-uri
-    // * requesting-user-name
-    request = ippNewRequest(operation);
-    httpAssembleURIf(HTTP_URI_CODING_ALL, uri, sizeof(uri), "ipp", "utf-8", "localhost",
-                     ippPort(), destination.toUtf8());
-    ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI, KCUPS_PRINTER_URI,
-                 "utf-8", uri);
-
-    return request;
-}
-
 QVariant KCupsConnection::ippAttrToVariant(ipp_attribute_t *attr)
 {
     QVariant ret;
@@ -975,11 +777,15 @@ QVariant KCupsConnection::ippAttrToVariant(ipp_attribute_t *attr)
     return ret;
 }
 
-bool KCupsConnection::retry(const char *resource)
+bool KCupsConnection::retry(const char *resource, int operation) const
 {
     ipp_status_t status = cupsLastError();
 
-    kDebug() << "cupsLastError():" << status << cupsLastErrorString();
+    if (operation != -1) {
+        kDebug() << ippOpString(static_cast<ipp_op_t>(operation)) << "last error:" << status << cupsLastErrorString();
+    } else {
+        kDebug() << operation << "last error:" << status << cupsLastErrorString();
+    }
 
     // When CUPS process stops our connection
     // with it fails and has to be re-established
@@ -993,7 +799,8 @@ bool KCupsConnection::retry(const char *resource)
 
         // Reconnect to CUPS
         if (httpReconnect(CUPS_HTTP_DEFAULT)) {
-            kWarning() << "IPP_INTERNAL_ERROR: failed to reconnect";
+            kWarning() << "Failed to reconnect" << cupsLastErrorString();
+
             // Server might be restarting sleep for a few ms
             msleep(500);
         }
@@ -1032,9 +839,9 @@ bool KCupsConnection::retry(const char *resource)
 
     if (forceAuth) {
         // force authentication
-        kDebug() << "cupsDoAuthentication() password_retries:" << password_retries;
+        kDebug() << "Calling cupsDoAuthentication() password_retries:" << password_retries;
         int ret = cupsDoAuthentication(CUPS_HTTP_DEFAULT, "POST", resource);
-        kDebug() << "cupsDoAuthentication() success:" << (ret == -1 ? true : false);
+        kDebug() << "Called cupsDoAuthentication(), success:" << (ret == -1 ? true : false);
 
         // If the authentication was succefull
         // sometimes just trying to be root works
